@@ -1,18 +1,28 @@
-import re
-from collections import OrderedDict
-from collections.abc import Mapping, Sequence, Set
 import datetime
 import json
 import os
 import pathlib
-from urllib.parse import parse_qs, quote, unquote, urlencode, urlparse, urlunparse
-from urllib.request import pathname2url, url2pathname
+import re
 import warnings
+from collections import OrderedDict
+from urllib.parse import (
+    parse_qs,
+    quote,
+    unquote,
+    urlencode,
+    urlparse,
+    urlunparse,
+)
+from urllib.request import pathname2url, url2pathname
 
+import requests
 # Third-Party
 import yaml
-from appdirs import user_cache_dir, user_data_dir
-import requests
+from platformdirs import user_cache_dir, user_data_dir
+
+
+def get_bool_from_env(var_name: str):
+    return os.getenv(var_name, '').lower() == 'true'
 
 
 APP_NAME = "Quilt"
@@ -24,9 +34,13 @@ TEMPFILE_DIR_PATH = BASE_PATH / "tempfiles"
 CONFIG_PATH = BASE_PATH / 'config.yml'
 OPEN_DATA_URL = "https://open.quiltdata.com"
 
-PACKAGE_NAME_FORMAT = r"([\w-]+/[\w-]+)(?:/(.+))?$"
+PACKAGE_NAME_FORMAT = r"^[\w-]+/[\w-]+$"
+DISABLE_TQDM = get_bool_from_env('QUILT_MINIMIZE_STDOUT')
+PACKAGE_UPDATE_POLICY = {'incoming', 'existing'}
+IS_CACHE_ENABLED = not get_bool_from_env('QUILT_DISABLE_CACHE')
 
-## CONFIG_TEMPLATE
+
+# CONFIG_TEMPLATE
 # Must contain every permitted config key, as well as their default values (which can be 'null'/None).
 # Comments are retained and added to local config, unless overridden by autoconfig via `api.config(<url>)`
 CONFIG_TEMPLATE = """
@@ -38,7 +52,7 @@ CONFIG_TEMPLATE = """
 # navigator_url: https://example.com
 navigator_url:
 
-# default_local_registry: <url string, default: local appdirs>
+# default_local_registry: <url string, default: local data directory (platform dependent)>
 # default target registry for operations like install and build
 default_local_registry: "{}"
 
@@ -65,22 +79,49 @@ apiGatewayEndpoint:
 # Binary API Gateway endpoint (e.g., for preview)
 binaryApiGatewayEndpoint:
 
+default_registry_version: 1
+
 """.format(BASE_PATH.as_uri() + '/packages')
 
+
+def get_pos_int_from_env(var_name):
+    val = os.getenv(var_name)
+    if val:
+        try:
+            val = int(val)
+        except ValueError:
+            val = None
+        if val is None or val <= 0:
+            raise ValueError(f'{var_name} must be a positive integer')
+        return val
+
+
 class QuiltException(Exception):
     def __init__(self, message, **kwargs):
         # We use NewError("Prefix: " + str(error)) a lot.
         # To be consistent across Python 2.7 and 3.x:
         # 1) This `super` call must exist, or 2.7 will have no text for str(error)
         # 2) This `super` call must have only one argument (the message) or str(error) will be a repr of args
-        super(QuiltException, self).__init__(message)
+        super().__init__(message)
         self.message = message
         for k, v in kwargs.items():
             setattr(self, k, v)
 
 
-class PhysicalKey(object):
-    __slots__ = ['bucket', 'path', 'version_id']
+class QuiltConflictException(QuiltException):
+    pass
+
+
+class RemovedInQuilt4Warning(FutureWarning):
+    pass
+
+
+class URLParseError(ValueError):
+    pass
+
+
+class PhysicalKey:
+    __slots__ = ('bucket', 'path', 'version_id')
 
     def __init__(self, bucket, path, version_id):
         """
@@ -108,7 +149,7 @@ class PhysicalKey(object):
 
         if parsed.scheme == 's3':
             if not parsed.netloc:
-                raise ValueError("Missing bucket")
+                raise URLParseError("Missing bucket")
             bucket = parsed.netloc
             assert not parsed.path or parsed.path.startswith('/')
             path = unquote(parsed.path)[1:]
@@ -117,24 +158,24 @@ class PhysicalKey(object):
             query = parse_qs(parsed.query)
             version_id = query.pop('versionId', [None])[0]
             if query:
-                raise ValueError(f"Unexpected S3 query string: {parsed.query!r}")
+                raise URLParseError(f"Unexpected S3 query string: {parsed.query!r}")
             return cls(bucket, path, version_id)
         elif parsed.scheme == 'file':
             if parsed.netloc not in ('', 'localhost'):
-                raise ValueError("Unexpected hostname")
+                raise URLParseError("Unexpected hostname")
             if not parsed.path:
-                raise ValueError("Missing path")
+                raise URLParseError("Missing path")
             if not parsed.path.startswith('/'):
-                raise ValueError("Relative paths are not allowed")
+                raise URLParseError("Relative paths are not allowed")
             if parsed.query:
-                raise ValueError("Unexpected query")
+                raise URLParseError("Unexpected query")
             path = url2pathname(parsed.path)
             if parsed.path.endswith('/') and not path.endswith(os.path.sep):
                 # On Windows, url2pathname loses the trailing `/`.
                 path += os.path.sep
             return cls.from_path(path)
         else:
-            raise ValueError(f"Unexpected scheme: {parsed.scheme!r}")
+            raise URLParseError(f"Unexpected scheme: {parsed.scheme!r}")
 
     @classmethod
     def from_path(cls, path):
@@ -145,7 +186,7 @@ class PhysicalKey(object):
             new_path = new_path.replace(os.path.sep, '/')
         # Add back a trailing '/' if the original path has it.
         if (path.endswith(os.path.sep) or
-            (os.path.altsep is not None and path.endswith(os.path.altsep))):
+                (os.path.altsep is not None and path.endswith(os.path.altsep))):
             new_path += '/'
         return cls(None, new_path, None)
 
@@ -235,7 +276,7 @@ def extract_file_extension(file_path_or_url):
 
 def read_yaml(yaml_stream):
     try:
-        if isinstance(yaml_stream, pathlib.PosixPath):
+        if isinstance(yaml_stream, pathlib.Path):
             with yaml_stream.open(mode='r') as stream:
                 return yaml.safe_load(stream)
         return yaml.safe_load(yaml_stream)
@@ -266,7 +307,7 @@ def write_yaml(data, yaml_path, keep_backup=False):
             path.parent.mkdir(parents=True)
         with path.open('w') as config_file:
             yaml.dump(data, config_file)
-    except Exception:     #! intentionally wide catch -- reraised immediately.
+    except Exception:     # intentionally wide catch -- reraised immediately.
         if backup_path.exists():
             if path.exists():
                 path.unlink()
@@ -298,7 +339,7 @@ def validate_url(url):
 class QuiltConfig(OrderedDict):
     def __init__(self, filepath, *args, **kwargs):
         self.filepath = pathlib.Path(filepath)
-        super(QuiltConfig, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def __setitem__(self, key, value):
         # Per chat in #engineering 4-5-19, strip navigator_url of trailing slash.
@@ -320,26 +361,11 @@ class QuiltConfig(OrderedDict):
         return "<{} at {!r} {}>".format(type(self).__name__, str(self.filepath), json.dumps(self, indent=4))
 
 
-def parse_sub_package_name(name):
-    """
-    Extract package name and optional sub-package path as tuple.
-    """
-    m = re.match(PACKAGE_NAME_FORMAT, name)
-    if m:
-        return tuple(m.groups())
-
-
 def validate_package_name(name):
     """ Verify that a package name is two alphanumeric strings separated by a slash."""
-    parts = parse_sub_package_name(name)
-    if not parts or parts[1]:
+    if not re.match(PACKAGE_NAME_FORMAT, name):
         raise QuiltException(f"Invalid package name: {name}.")
 
-def get_package_registry(path=None):
-    """ Returns the package registry root for a given path """
-    if path is None:
-        path = get_from_config('default_local_registry')
-    return path.rstrip('/') + '/.quilt'
 
 def configure_from_url(catalog_url):
     """ Read configuration settings from a Quilt catalog """
@@ -368,24 +394,24 @@ def configure_from_url(catalog_url):
 
     # Use our template + their configured values, keeping our comments.
     for key, value in new_config.items():
-        if not key in config_template:
+        if key not in config_template:
             continue
         config_template[key] = value
     write_yaml(config_template, CONFIG_PATH, keep_backup=True)
     return config_template
 
+
 def config_exists():
     """
     Returns True if a config file (config.yml) is installed.
     """
     return CONFIG_PATH.exists()
 
+
 def user_is_configured_to_custom_stack():
-    """Look at the users stack to see if they have configured to their own stack. There is currently no way to
-    distinguish between someone who has not configured their stack and someone who has intentionally configured
-    their stack to use open.quiltdata.com"""
+    """Look at the users stack to see if they have configured to their own stack."""
     configured_nav_url = get_from_config("navigator_url")
-    return configured_nav_url is not None and configured_nav_url != OPEN_DATA_URL
+    return configured_nav_url is not None
 
 
 def configure_from_default():
@@ -406,6 +432,7 @@ def configure_from_default():
         local_config = config_template
     return local_config
 
+
 def load_config():
     """
     Read the local config using defaults from CONFIG_TEMPLATE.
@@ -415,21 +442,25 @@ def load_config():
         local_config.update(read_yaml(CONFIG_PATH))
     return local_config
 
+
 def get_from_config(key):
     return load_config().get(key)
 
+
 def get_install_location():
     loc = get_from_config('default_install_location')
     if loc is None:
         loc = get_from_config('default_local_registry').rstrip('/')
     return loc
 
+
 def set_config_value(key, value):
     # Use local configuration (or defaults)
     local_config = load_config()
     local_config[key] = value
     write_yaml(local_config, CONFIG_PATH)
 
+
 def quiltignore_filter(paths, ignore, url_scheme):
     """Given a list of paths, filter out the paths which are captured by the
     given ignore rules.
@@ -471,6 +502,7 @@ def quiltignore_filter(paths, ignore, url_scheme):
     else:
         raise NotImplementedError
 
+
 def validate_key(key):
     """
     Verify that a file path or S3 path does not contain any '.' or '..' separators or files.
@@ -487,6 +519,7 @@ def validate_key(key):
                 f"A package entry key cannot contain a file or folder named '.' or '..' in its path."
             )
 
+
 def catalog_s3_url(catalog_url, s3_url):
     """
     Generate a URL to the Quilt catalog page for an object in S3
@@ -509,10 +542,12 @@ def catalog_s3_url(catalog_url, s3_url):
             url += f"?{urlencode(params)}"
     return url
 
-def catalog_package_url(catalog_url, bucket, package_name, package_timestamp="latest"):
+
+def catalog_package_url(catalog_url, bucket, package_name, package_timestamp="latest", tree=True):
     """
     Generate a URL to the Quilt catalog page of a package. By default will go to the latest version of the package,
     but the user can pass in the appropriate timestamp to go to a different version.
+    Disabling tree by passing `tree=False` will generate a package URL without tree path.
 
     Note: There is currently no good way to generate the URL given a specific tophash
     """
@@ -520,7 +555,7 @@ def catalog_package_url(catalog_url, bucket, package_name, package_timestamp="la
     assert package_name is not None, "The package_name parameter must not be None"
     validate_package_name(package_name)
 
-    return f"{catalog_url}/b/{bucket}/packages/{package_name}/tree/{package_timestamp}"
-
-
-
+    package_url = f"{catalog_url}/b/{bucket}/packages/{package_name}"
+    if tree:
+        package_url = package_url + f"/tree/{package_timestamp}"
+    return package_url
